#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/select.h>

typedef int t_pipe[2];

/* struktura wiadomosci */
enum t_sender {current=0, left, right, prev, head};
enum t_content {no_operation = 0, 
		make_step    = 1, 
		step_done    = 2, 
		step_fail    = 4, 
		left_step    = 8, 
		right_step   = 16, 
		call_exit    = 32};

struct msg {
  enum t_sender sender;
  enum t_content content;
};

typedef struct msg t_msg;

#define MSGSIZE sizeof(t_msg)

/* noga, leg - prawa, albo lewa, rd i wr - lacza do wlasciwego czlonu */
void noga(enum t_sender leg, int rd, int wr) {
  t_msg msg={current, no_operation};
  srand(getpid()*getppid()*time(NULL));
  while (msg.content!=call_exit) {
    if (msg.content==make_step) {
      usleep((int)(3000000.0*rand()/(RAND_MAX+1.0)));
      msg.sender=leg;
      msg.content=step_done|((leg==left) ? left_step : right_step);
      write(wr, &msg, MSGSIZE);
    }
    read(rd, &msg, MSGSIZE);
  }
  close(rd); close(wr);
  exit(0);
}

/* czlon n, czyta lacza prd i crd, zapisuje do out* */
void czlon(int n, int prd, int crd, int outl, int outr, int outn, int outp) {
  t_msg msg={current, no_operation};
  int s, i, left=0, right=0;
  fd_set set;
  struct timeval t;
  FD_ZERO(&set);

  while (msg.content!=call_exit) {
    do { /* czekaj, az select stwierdzi ze jest wiadomosc w prd lub crd */
      FD_SET(prd, &set);
      FD_SET(crd, &set);
      t.tv_sec = 1;
      t.tv_usec = 0; 
    } while (select(crd+1, &set, NULL, NULL, &t)!=1);
    if (FD_ISSET(crd, &set)) s=read(crd, &msg, MSGSIZE);
    else s=read(prd, &msg, MSGSIZE);

    /* przetworzenie wiadomosci */
    if (msg.sender==head) { /* wiadomosc od glowy, przekaz nastepnemu i nogom */
      left=right=0; /* zresetuj stan nog poprzednich */
      write(outl, &msg, MSGSIZE);
      write(outr, &msg, MSGSIZE);
      write(outn, &msg, MSGSIZE);
    }

    if (msg.content&step_done&&msg.sender!=prev) { /* potwierdzenie kroku od nogi */
      msg.sender=prev; /* przygotowanie potwierdzenia dla poprzedniego czlonu */
      if ((msg.content&left_step && left==0)||
	  (msg.content&right_step && right==0)) {
	printf("czlon %d: %s krok\n", n, msg.content&left_step ? "lewy" : "prawy");
      } else {
	printf("czlon %d: %s potknicie\n", n, msg.content&left_step ? "lewy" : "prawy");
	msg.content^=step_done|step_fail; /* wykasuj step_done, ustaw step_fail */
      }
      write(outp, &msg, MSGSIZE); /* przekaz poprzedniemu */
    }

    if (msg.sender==prev) { /* krok poprzedniego czlonu wykonany */
      if (msg.content&left_step) left++;
      else right++;
    }
  }
  close(crd); close(prd); close(outl); close(outp); close(outn); close(outp);
  exit(0);
}

/* glowa (wyk. w procesie glownym), rd i wr - lacza do pierwszego czlonu */
void glowa(int rd, int wr) {
  char c;
  t_msg msg={head, make_step};
  printf("polecenia:\nk - krok\nq - wyjcie\n");
  while ((c=getchar())!='q') {
    if (c=='k') write(wr, &msg, MSGSIZE);
  }
  msg.content=call_exit;
  write(wr, &msg, MSGSIZE);
  close(rd); close(wr);
}

/* utworzenie lacz. prev_rd i prev_wr - lacza do poprzedniego czlonu do read i write */
/* n - liczba laczy do utworzenia, cn - numer aktualnego czlonu*/
void dodaj_czlony(int n, int cn, int prev_rd, int prev_wr) {
  int i;
  t_pipe curr_out[3];
  t_pipe curr_in;
  for (i=0; i<3; i++) pipe(curr_out[i]);
  pipe(curr_in);
  if (fork()==0) {
    close(curr_out[0][1]);
    close(curr_in[0]);
    noga(left, curr_out[0][0], curr_in[1]);
  }
  if (fork()==0) {
    close(curr_out[1][1]);
    close(curr_in[0]);
    noga(right, curr_out[1][0], curr_in[1]);
  }
  if (n>1)
    if (fork()==0) {
      close(curr_out[2][1]);
      close(curr_in[0]);
      dodaj_czlony(n-1, ++cn, curr_out[2][0], curr_in[1]);
      exit(0);
    }
  close(curr_in[1]);
  for (i=0; i<3; i++) 
    close(curr_out[i][0]);
  czlon(cn, prev_rd, curr_in[0], curr_out[0][1], curr_out[1][1], curr_out[2][1], prev_wr);
}

int main(int argc, char *argv[]) {
  int n;
  t_pipe rd, wr;
  if (argc<2) {
    printf("stonoga n_czlonow\n"); return 1;
  }
  n=atoi(argv[1]);
  pipe(rd);
  pipe(wr);
  if (fork()==0) {
    close(rd[1]);
    close(wr[0]);
    dodaj_czlony(n, 1, rd[0], wr[1]);
  }
  close(rd[0]);
  close(wr[1]);
  glowa(wr[0], rd[1]);
  return 0;
}
