#include <stdio.h>
#include <stdlib.h> /* RAND_MAX */
#include <time.h> /* struct timespec */
#include <linux/ipc.h>
#include <linux/sem.h>
#include <linux/msg.h>
#include <linux/shm.h>
#include "struktury.h"

int pocztaid;

typedef struct wiad_kol {
  long nr_skrzynki;
  wiadomosc_t wiadomosc;
} wiad_kol_t;

void czekaj_chwile(int maks_sec) {
  struct timespec czas = {0, 0};
  czas.tv_sec = (int) (maks_sec*1.0*rand()/(RAND_MAX*1.0));
  nanosleep(&czas, NULL);
}

void zadaj_dostep(int sem_id) {
  struct sembuf s[3] = {{SEM_LL, 0, 0}, {SEM_LL, +1, 0}, {SEM_LK, 0, 0}};
  /* czekaj az liczba listonoszy=0, zwieksz liczbe listonoszy, czekaj az liczba klientow=0 */
  semop(sem_id, s, 3);
}

void zwolnij_dostep(int sem_id) {
  struct sembuf s = {SEM_LL, -1, 0}; /* zmniejsz liczbe listonoszy */
  semop(sem_id, &s, 1);
}

int w_partii(int partia) { /* ile jest wiadomosci w kolejce partia */
  struct msqid64_ds tmp;
  if (msgctl(partia, IPC_STAT, &tmp) == -1)
    perror("w_partii");
  else return tmp.msg_qnum;
}

int z_pliku_do_partii(FILE *plik, int partia) { 
/* skopiuj wiadomosci z pliku do partii, zakoncz, jesli napotkano koniec partii (nowa linia),
   lub napotkano koniec pliku. jesli nie napotkano konca pliku zwraca 1 (w przyszlosci bedzie
   jeszcze trzeba czytac z pliku bo cos tam jest), jesli napotkano koniec pliku zwraca 0  */
  char buf[144];
  wiad_kol_t wiad;
  int i;
  while ((!feof(plik)) && (fgets(buf, 144, plik)!=NULL) && (strlen(buf)>1)) {
    printf("Podniosem wiadomo\n");
    memset(&wiad.wiadomosc, 0, sizeof(wiadomosc_t));
    sscanf(buf, "%d %[ -~]", &wiad.nr_skrzynki, &wiad.wiadomosc);
    wiad.nr_skrzynki++; /* wiadomosci w kolejce maja adres +1, zeby moc zaadresowac skrzynke nr. 0 */
    if (msgsnd(partia, &wiad, sizeof(wiadomosc_t), 0) == -1) { /* moze sie zdarzyc zapchanie partii */
      perror("z_pliku_do_pliku msgsnd");
    }
  }
  return !feof(plik);
}

int z_partii_do_partii(int zpartia, int dopartia) { /* przenies wiadomosci z zpartia do dopartia */
  wiad_kol_t wiad;
  while (msgrcv(zpartia, &wiad, sizeof(wiadomosc_t), 0, IPC_NOWAIT) != -1) {
    msgsnd(dopartia, &wiad, sizeof(wiadomosc_t), 0);
  }
  return TRUE;
}

int roznies_partie(skrzynka_t* skrzynki, int lskrzynek, int pojemnosc, int partia, int odlozone, int sem_id) {
  /* roznies wiadomosci do skrzynek z kolejki partia. jesli skrzynki zajete, przesylaj wiadomosci
     do kolejki odlozone. gdy potrzebujesz dostep do skrzynek, uzywaj semafora sem_id */
  wiad_kol_t wiad;
  wiadomosc_t* adr;
  while (msgrcv(partia, &wiad, sizeof(wiadomosc_t), 0, IPC_NOWAIT) != -1) { /* poki sa wiadomosci */
    wiad.nr_skrzynki--; /* patrz komentarz: z_pliku_do_partii() o numeracji skrzynek w kolejkach */
    zadaj_dostep(sem_id);
    if (wiad.nr_skrzynki<lskrzynek && (skrzynki[wiad.nr_skrzynki].uzywana)) { /* jest tyle skrzynek i akurat ta jest uzywana */
      if (skrzynki[wiad.nr_skrzynki].lwiadomosci<pojemnosc) { /* jesli jest jeszcze miejsce, wstawiamy do skrzynki */
	printf("Wiadomo do skrzynki %d\n", wiad.nr_skrzynki);
	adr = (wiadomosc_t*) ((int)skrzynki + (int)skrzynki[wiad.nr_skrzynki].wiadomosci);
	memcpy(adr[skrzynki[wiad.nr_skrzynki].lwiadomosci++], wiad.wiadomosc, 128);
      } else { /* jesli nie ma miejsca, przenosimy do odlozonych */
	printf("Skrzynka %d: brak miejsca, przel pniej\n", wiad.nr_skrzynki);
	wiad.nr_skrzynki++;
	msgsnd(odlozone, &wiad, sizeof(wiadomosc_t), 0);
      }
      zwolnij_dostep(sem_id);
      czekaj_chwile(5);
    }
    else {
      zwolnij_dostep(sem_id);
      printf("Ehh, znowu zy nadawca, skrzynka %d nie jest uywana\n", wiad.nr_skrzynki);
    }
  }
}

int main(int argc, char *argv[]) {
  int         ta_partia, nast_partia; /* partia aktualnie roznoszona i partia odlozonych */
  int         skr_id, sem_id; /* id pamieci wspoldzielonej, id semaforow */
  skrzynka_t* skrzynki; /* wskaznik do pamieci ze skrzynkami */
  int         lskrzynek, pojemnosc; /* liczba skrzynek i ich pojemnosc */
  FILE*       listy; /* plik z listami */

  if (argc<2) {
    printf("Uywanie: listonosz plik\n");
    return APP_FAILED;
  }  

  srand(time(NULL)*getpid());

  pocztaid = ftok(SCIEZKAPOCZTA, (int) 'a');
  if (! poczta_otworz(pocztaid)) {
    printf("Nie mona si dosta do poczty\n");
    return APP_FAILED;
  }
  
  if ((listy = fopen(argv[1], "r")) == NULL) {
    perror("fopen");
    return APP_FAILED;
  }

  skr_id = zapytaj(NADAWCA_LISTONOSZ, OPER_ID_SEGMENTU, 0);
  pojemnosc = zapytaj(NADAWCA_LISTONOSZ, OPER_ILE_MIEJSCA, 0);
  lskrzynek = zapytaj(NADAWCA_LISTONOSZ, OPER_ILE_SKRZYNEK, 0);

  sem_id = semget(pocztaid, 0, 0660);
  if ((int)(skrzynki = (skrzynka_t*) shmat(skr_id, NULL, 0660)) == -1) {
    perror("shmat()");
    return APP_FAILED;
  }

  if ((ta_partia = msgget(IPC_PRIVATE, 0660)) == -1) {
    perror("msgget ta_partia");
    return APP_FAILED;
  }
  
  if ((nast_partia = msgget(IPC_PRIVATE, 0660)) == -1) {
    perror("msgget nast_partia");
    return APP_FAILED;
  }

  while (z_pliku_do_partii(listy, ta_partia) || w_partii(ta_partia)) {
    /* dopoki sa wiadomosci w pliku lub w partii wykonuj swe zadanie */
    printf("Id roznie parti listw\n");
    roznies_partie(skrzynki, lskrzynek, pojemnosc, ta_partia, nast_partia, sem_id);
    z_partii_do_partii(nast_partia, ta_partia);
    printf("Odpoczywam po rozniesieniu partii\n");
    czekaj_chwile(5);
  }

  msgctl(ta_partia, IPC_RMID, 0);
  msgctl(nast_partia, IPC_RMID, 0);
  poczta_zamknij();
  return APP_SUCCESS;
}
