
#ifndef DRIVER
#define DRIVER

#include "defs.h"

class driver{
public: //interface
 driver(int lpt=1, int irq=7);
 ~driver();
 void stop();
 void move(double s);
 void rotate(double angle);
 void set_automode();
 void play(int frequency, double duration);
private: //control
 bool automode; //flaga wlaczenia autopilota
 enum{none,forward,backward,left_rot,right_rot} last_move;
 int s_left, s_right; //droga do przebycia dla lewego i prawego kola (w korkach)
 void step(); //funkcja wywolywana przy przerwaniu zegarowym
 void autopilot();
 inline void step_left(int dir); //zrob krok lewym/prawym kolem
 inline void step_right(int dir);
private: //handling interrupts
 bool installed;
 bool install(int lpt, int irq);
 void uninstall();
 static driver *current;
private:
 word lpt_base; //adres lpt'a
 word pic_base; //adres kontrolera przerwan (PIC - programmable interrupt controller)
 byte lpt_int; //numer przerwania lpt'a
 byte irq_mask; //maska przerwania w OCW1 kontrolera
 static void interrupt play_int(...);
 static void interrupt clock_int_handler(...);
 static void interrupt lpt_int_handler(...);
 struct{
  void interrupt (*old_clock_int)(...);
  void interrupt (*old_lpt_int)(...);
 } backup;
private: //motor control
 enum{cycle=4};
 int left,right; //pozycja lewego i prawego silnika
 static const byte left_motor[cycle];
 static const byte right_motor[cycle];
private:
 byte last_status; //bajt statusowy (po konwersji)
 byte control; //bajt kontrolny
 int piezo_t; //ustawiony czas brzeczyka
};

#endif
