#include <stdio.h>
#include <stdlib.h> /* RAND_MAX */
#include <time.h> /* struct timespec */
#include <linux/ipc.h>
#include <linux/sem.h>
#include <linux/msg.h>
#include <linux/shm.h>
#include "struktury.h"

int pocztaid;

void czekaj_chwile(int maks_sec) {
  struct timespec czas = {0, 0};
  czas.tv_sec = (int) (maks_sec*1.0*rand()/(RAND_MAX*1.0));
  nanosleep(&czas, NULL);
}

void zadaj_dostep(int sem_id) {
  struct sembuf s[2] = {{SEM_LL, 0, 0}, {SEM_LK, +1, 0}};
  /* czekaj, az liczba listonoszy=0, zwieksz liczbe klientow o 1 */
  semop(sem_id, s, 2);
}

void zwolnij_dostep(int sem_id) {
  struct sembuf s = {SEM_LK, -1, 0}; /* zmniejsz liczbe klientow o 1 */
  semop(sem_id, &s, 1);
}

int czytaj_listy(skrzynka_t* skrzynka, wiadomosc_t* wiadomosci, int sem_id) {
  zadaj_dostep(sem_id);
  if (skrzynka->lwiadomosci>0) {
    do {
      printf("%s\n", wiadomosci[--skrzynka->lwiadomosci]);
    } while (skrzynka->lwiadomosci>0);
  } else
    printf("Brak wiadomoci\n");
  zwolnij_dostep(sem_id);
}

int main(int argc, char *argv[]) {
  int          skr_id, sem_id; /* identyfikatory pamieci wspoldzielonej i semaforow */
  int          lprob, nr; /* liczba prob (podawana jako parametr), numer skrzynki */
  int          cierpliwosc=10; /* cierpliwosc - ile razy probowac sie laczyc */
  skrzynka_t*  skrzynka; /* wskaznik do wlasnej skrzynki w obszarze pam.wspoldz. */
  wiadomosc_t* wiadomosci; /* bezwzgl. wskaznik do wiadomosci w skrzynce */
  if (argc<2) {
    printf("Uywanie: klient lprob\n");
    return APP_FAILED;
  }  

  if ((lprob = atol(argv[1])) == 0) {
    printf("lprob musi by liczb >0\n");
    return APP_FAILED;
  }

  srand(time(NULL)*getpid());

  pocztaid = ftok(SCIEZKAPOCZTA, (int) 'a');

  while (--cierpliwosc && (! poczta_otworz(pocztaid))) {
    printf("Nie mona si dosta do poczty...\n");
    czekaj_chwile(3);
  }

  if (! cierpliwosc) {
    printf("Poczta nie dostpna\n");
    return APP_FAILED;
  }

  if ((nr = zapytaj(NADAWCA_KLIENT, OPER_NOWA_SKRZYNKA, 0)) != -1) {
    printf("Mam skrzynk %d\n", nr);
  } else {
    printf("Nie mam skrzynki\n");
    poczta_zamknij();
    return APP_FAILED;
  }
  skr_id = zapytaj(NADAWCA_KLIENT, OPER_ID_SEGMENTU, 0);
  poczta_zamknij();

  sem_id = semget(pocztaid, 0, 0660);
  if ((int)(skrzynka = (skrzynka_t*) shmat(skr_id, NULL, 0660)) == -1) {
    perror("shmat()");
    return APP_FAILED;
  }
  /* wskazniki do skrzynek w pamieci wspoldzielonej liczone sa od poczatkowego adresu tej pamieci,
     czyli wzglednie. zaleznie od zmapowania pam.wspoldz. kazdy program zmienia sobie wskaznik do
     wiadomosci o odp. przesuniecie (adres skrzynek) */
  wiadomosci = (wiadomosc_t*) ((int)skrzynka + (int)skrzynka[nr].wiadomosci);
  skrzynka = &skrzynka[nr];

  while (lprob--) {
    printf("Id na poczt...\n");
    czytaj_listy(skrzynka, wiadomosci, sem_id);
    printf("Id do domu\n");
    czekaj_chwile(5);
  }

  while ((! poczta_otworz(pocztaid))) {
    printf("Nie mona si dosta do poczty...\n");
    czekaj_chwile(3);
  }

  zapytaj(NADAWCA_KLIENT, OPER_USUN_SKRZYNKE, nr);
  poczta_zamknij();
  return APP_SUCCESS;
}
