#include <stdio.h>
#include <sys/types.h> /* key_t */
#include <linux/ipc.h>
#include <linux/shm.h>
#include <linux/sem.h>
#include <linux/msg.h>
#include <unistd.h>
#include <errno.h>
#include <stdarg.h> /* va_list ... */
#include <time.h> /* sizeof struct tm */
#include <signal.h> /* SIGINT */
#include "struktury.h"

#define ROZM_SKRZYNEK lskrzynek*(sizeof(skrzynka_t)+(pojemnosc*sizeof(wiadomosc_t)))

int         lskrzynek=0;    /* liczba skrzynek dostepnych na poczcie */
int         pojemnosc=0;    /* ile wiadomosci pomiesci skrzynka */
key_t       pocztaid;       /* key dla funkcji IPC */
skrzynka_t* skrzynki;       /* przestrzen wspoldzielona zawierajaca skrzynki */
int         skrzynki_wolne; /* kolejka z wolnymi skrzynkami */
int         skr_id;         /* id segmentu pamieci wspoldzielonej */
int         sem_id;         /* id semaforow IPC */
int         kolejka;        /* kolejka z zapytaniami klientow/listonoszy */

void loguj(char *format, ...) {
  va_list   ap;
  time_t    teraz_sec = time(NULL);
  struct tm teraz;
  va_start(ap, format);
  gmtime_r(&teraz_sec, &teraz);
  fprintf(stderr, "poczta [%d-%02d-%02d %02d:%02d:%02d]: ", 1900+teraz.tm_year, teraz.tm_mon, 
	                                         teraz.tm_mday, teraz.tm_hour, 
                                                 teraz.tm_min, teraz.tm_sec);
  vfprintf(stderr, format, ap);
  va_end(ap);
}

void zadaj_dostepu(void) {
  struct sembuf s[3] = {{SEM_LL, 0, 0}, {SEM_LL, +1, 0}, {SEM_LK, 0, 0}};
  /* poczta potrzebuje dostep do zapisu, wiec ustawia semafory jakby byla listonoszem */
  semop(sem_id, s, 3);
}

void zwolnij_dostep(void) {
  struct sembuf s = {SEM_LL, -1, 0};
  semop(sem_id, &s, 1);
}

int rejestruj_skrzynke(void) { /* rejestruje skrzynke z kolejki wolnych i zwraca jej numer, lub -1 jesli brak skrzynek */
  long              num;
  struct msqid64_ds buf;
  msgctl(skrzynki_wolne, IPC_STAT, &buf);
  if (buf.msg_qnum==0) 
    return -1;

  if (msgrcv(skrzynki_wolne, &num, 0, 0, 0) == -1) perror("rejestruj_skrzynke");
  num--; // patrz: num++ w zwolnij_skrzynke
  zadaj_dostepu();
  skrzynki[num].uzywana = 1;
  zwolnij_dostep();
  loguj("Zarejestrowano skrzynk %d\n", num);
  return num;
}

int zwolnij_skrzynke(int num) { /* zwalnia skrzynke o numerze num i dodaje do kolejki wolnych */
  loguj("Zwalniam skrzynk %d\n", num);
  zadaj_dostepu();
  skrzynki[num].uzywana = 0;
  skrzynki[num].lwiadomosci = 0;
  zwolnij_dostep();
  num++; // w kolejce przechowujemy numery skrzynek+1, bo 0 jest niepoprawne!
  if (msgsnd(skrzynki_wolne, &num, 0, 0) == -1) perror("zwolnij_skrzynke()");
  num--;
  return 0;
}

void serwuj(int kolejka) {
  int oczekujace, loczekujacych=0; /* oczekujace - kolejka komunikatow, ktorych nie udalo sie wyslac */
  int tmp;
  zapytanie_t wiad, odp;
  oczekujace = msgget(IPC_PRIVATE, 0660);
  odp.nadawca = NADAWCA_POCZTA;
  odp.adr_zwrotny = kolejka;

  errno=0;
  while ((msgrcv(kolejka, &wiad, ROZM_TRESCI, 0, 0) != -1)&&(errno!=EINTR)) { 
    /* tak dlugo jak nie trafisz na blad, lub nie otrzymasz sygnalu konca (SIGINT) lub innego przerwania */
    odp.operacja = wiad.operacja;
    loguj("operacja (OD:%d KOD:%d, ADR_ZWR:%d)\n", wiad.nadawca, wiad.operacja, wiad.adr_zwrotny);

    switch (wiad.operacja) {
    case (OPER_NOWA_SKRZYNKA):
      odp.numer = rejestruj_skrzynke();
      break;
    case (OPER_USUN_SKRZYNKE):
      odp.numer = zwolnij_skrzynke(wiad.numer);
      break;
    case (OPER_ILE_MIEJSCA):
      odp.numer =  pojemnosc;
      break;
    case (OPER_ILE_SKRZYNEK):
      odp.numer = lskrzynek;
      break;
    case (OPER_ID_SEGMENTU):
      odp.numer = skr_id;
      break;
    }
    if (msgsnd(wiad.adr_zwrotny, &odp, ROZM_TRESCI, IPC_NOWAIT) == -1) { /* jesli nie dalo sie wyslac */
      switch (errno) {
      case (EAGAIN): /* brak miejsca w kolejce adresata, dodaj do kolejki oczekujacych*/
	odp.adr_zwrotny = wiad.adr_zwrotny;
	msgsnd(oczekujace, &odp, ROZM_TRESCI, 0);
	loczekujacych++;
	break;
      case (EIDRM): /* adresat usunal kolejke, przed odebraniem potwierdzenia nadania - usun mu skrzynke */
	if (wiad.operacja == OPER_NOWA_SKRZYNKA) {
	  zwolnij_skrzynke(odp.numer);
	}
	break;
      }
      loguj("odlozone (DO:%d, KOD:%d)\n", wiad.adr_zwrotny, odp.operacja);
    }
    if (loczekujacych>0) { /* jesli jest cos w kolejce oczekujacych */
      msgrcv(oczekujace, &odp, ROZM_TRESCI, 0, 0); /* odbierz to */
      tmp = odp.adr_zwrotny;
      odp.adr_zwrotny = kolejka;
      loguj("wysylam odlozone (DO:%d, KOD:%d)\n", tmp, odp.operacja);
      msgsnd(tmp, &odp, ROZM_TRESCI, 0); /* i wyslij ponownie, blokujac sie jesli tym razem znow zajete */
      loczekujacych--;
    }
  }

  if (errno!=EINTR) { /* wystapil blad inny niz signal -> nie przewidziany */
    loguj("Blad odczytu kolejki\n");
    perror("msgrcv()");
  }
  msgctl(oczekujace, IPC_RMID, NULL);
}

int zajmij_zasoby(void) {
  int i;
  if ((skr_id = shmget(pocztaid, ROZM_SKRZYNEK, IPC_CREAT|0660)) == -1) {
    perror("shmget()");
    return FALSE;
  }
  if ((skrzynki = (skrzynka_t*) shmat(skr_id, NULL, 0)) == (void*)-1) {
    perror("shmat()");
    return FALSE;
  }
  if ((sem_id = semget(pocztaid, 3, IPC_CREAT|IPC_EXCL|0660)) == -1) {
    perror("semget()");
    return FALSE;
  }
  if (semctl(sem_id, SEM_LS, SETVAL, lskrzynek) == -1) 
    perror("nie");
  if ((kolejka = msgget(pocztaid, IPC_CREAT|IPC_EXCL|0660)) == -1) {
    perror("msgget()");
    return FALSE;
  }
  if ((skrzynki_wolne = msgget(IPC_PRIVATE, 0660)) == -1) {
    perror("msgget()");
    return FALSE;
  }

  for (i=0; i< lskrzynek; i++) { 
    /* ustaw wskazniki do wiadomosci <-> sformatuj przestrzen skrzynek. 
       UWAGA adres wiadomosci wzgledny, liczony od poczatku pamieci wspoldzielonej */
    skrzynki[i].wiadomosci = (wiadomosc_t*) ((lskrzynek*sizeof(skrzynka_t)) + i*(pojemnosc*sizeof(wiadomosc_t)));
    zwolnij_skrzynke(i); /* dodaje skrzynke do kolejki wolnych */
  }
  return TRUE;
}

void zwolnij_zasoby(void) {
  if (shmdt(skrzynki) == -1) {perror("shmdt()"); };
  if (shmctl(skr_id, IPC_RMID, NULL) == -1) {perror("shmctl(skr_id rm)"); };
  if (semctl(sem_id, 0, IPC_RMID, 0) == -1) {perror("semctl(sem rm)"); };
  if (msgctl(kolejka, IPC_RMID, NULL) == -1) {perror("msgctl(kol_id, rm)"); };
  if (msgctl(skrzynki_wolne, IPC_RMID, NULL) == -1) {perror("msgctl(skrzynki_wolne, rm)"); };
  loguj("Zasoby zwolnione\n");
}

void bezpieczne_wyjscie(int sygnal) { /* wrapper sygnau */
  loguj("Otrzymano sygna %d\n", sygnal);
  zwolnij_zasoby();
}

int main(int argc, char *argv[]) {
  if (argc<3) {
    printf("Uywanie: poczta lskrzynek pojemnosc\n");
    return APP_FAILED;
  }
  if ((lskrzynek=atoi(argv[1]))==0) {
    printf("Liczba skrzynki musi by liczb >0\n");
    return APP_FAILED;
  } 
  if ((pojemnosc=atoi(argv[2]))==0) {
    printf("Pojemno musi by liczb >0\n");
    return APP_FAILED;
  }  

  pocztaid = ftok(SCIEZKAPOCZTA, (int) 'a');
  if (! zajmij_zasoby()) {
    return APP_FAILED;
  }
  signal(SIGINT, bezpieczne_wyjscie); 
  loguj("Poczta otwarta (SIGINT aby zamkn)\n");
  
  serwuj(kolejka);

  if (errno!=EINTR) {
    loguj("Zwyczajne wyjcie\n");
    zwolnij_zasoby();
  }
  return APP_SUCCESS;
}
