#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <stdarg.h>

#include "struktury.h"

#define BUFSIZE 1024

/* typedef enum tryb: sposob wykonania funkcji: 
     - nie wykonywac, tylko zapamietac z jakimi zostala wywolana argumentami;
     - wykonac z takimi argumentami, z jakimi zostala wywolana ostatnio;
     - wykonac z takimi argumentami, jakie podano w biezacym wywolaniu;
*/
typedef enum tryb {zapamietaj_ustawienia, wykonaj, wykonaj_dla_argumentow} tryb_t;
typedef enum operacja_kasy {KASA_STAN='=', KASA_KP='+', KASA_KW='-'} operacja_kasy_t;

void logowanie(char *format, ...) {
  va_list ap;
  time_t teraz_sec = time(NULL);
  struct tm teraz;
  va_start(ap, format);
  gmtime_r(&teraz_sec, &teraz);
  fprintf(stderr, "sklep [%d-%02d-%02d %02d:%02d:%02d]: ", 1900+teraz.tm_year, teraz.tm_mon, 
	                                         teraz.tm_mday, teraz.tm_hour, 
                                                 teraz.tm_min, teraz.tm_sec);
  vfprintf(stderr, format, ap);
  va_end(ap);
}

int kasa(int kwota, operacja_kasy_t operacja) {
  static int stan_kasy=0;
  switch (operacja) {
  case (KASA_KP):
    stan_kasy+=kwota;
    break;
  case (KASA_KW):
    stan_kasy-=kwota;
    break;
  }
  logowanie("Stan kasy: %d (%c)\n", stan_kasy, operacja);
  return stan_kasy;
}

void nabij_na_kase(towar_t towar) {
  logowanie("Zakup \"%s\" za: %d. Na magazynie: %d\n", towar.nazwa, towar.cena, towar.ilosc);
  kasa(towar.cena, KASA_KP);
}

towar_t* laduj_towar(char *plik, int *l_towarow) {
  char buf[BUFSIZE]={0};
  int towarow_do_zaladowania=0, nr_towaru=0;
  towar_t* towary=NULL;
  FILE *p;
  if ((p = fopen(plik, "r")) == NULL)
    return NULL;
  fgets(buf, BUFSIZE, p);
  towarow_do_zaladowania = *l_towarow = atoi(buf);
  if ((towary = (towar_t*) malloc(towarow_do_zaladowania*sizeof(towar_t))) == NULL)
    return NULL;
  while (towarow_do_zaladowania-- > 0) {
    fgets(buf, BUFSIZE, p);
    towary[nr_towaru].nazwa = (char*) malloc((int)memchr(buf, ' ', BUFSIZE)-(int)buf+1);
    if (towary[nr_towaru].nazwa==NULL) return NULL;
    /* dlugosc ciagu dla pola nazwa = adres bufora - adres pierwszej spacji w buforze + 1bajt na \0 */
    sscanf(buf, "%s %d %d", towary[nr_towaru].nazwa, &towary[nr_towaru].cena, &towary[nr_towaru].ilosc);
    nr_towaru++;
  }
  return towary;
}

void wypisz_towary(int lacze, towar_t* towar, int l_towarow) {
  char buf[23]={0};
  int k;
  for (k=0; k<l_towarow; k++) {
    sprintf(buf, " %-14s %5d\n", towar[k].nazwa, towar[k].cena); 
    /* uwaga, dlugosc nie wieksza, niz rozmiar buf */
    write(lacze, buf, strlen(buf));
  }
}

int sprzedaj_towar(char *nazwa, towar_t* magazyn, int l_towarow) {
  int i, dl_naz, dl;
  dl_naz=strlen(nazwa);
  for (i=0; i<l_towarow; i++)
    if ((strncmp(nazwa, magazyn[i].nazwa, dl_naz<strlen(magazyn[i].nazwa) ? dl_naz : strlen(magazyn[i].nazwa))==0) && 
	(magazyn[i].ilosc>0)) { 
      magazyn[i].ilosc--;
      nabij_na_kase(magazyn[i]);
      return 1;
    }
  return 0;
}

int dodaj_klienta(char* nowy) {
  int id;
  if ((id = open(nowy, O_WRONLY)) == -1)
    return 0;
  else 
    return id;
}

void oferuj(int lacze, towar_t* towar, int l_towarow) {
  komunikat_t zapytanie;
  komunikat_arg_t argument={0, NULL};
  char argument_const[100]={0};
  char tmp;
  int k;
  while (read(lacze, &zapytanie, sizeof(komunikat_t))) {
    if (zapytanie.operacja & (OP_NOWE_POL|OP_KUP) ) { /* pobierz argument */
      read(lacze, &argument.dlugosc, sizeof(char));
      memset(argument_const, 0, 100);
      read(lacze, argument_const, argument.dlugosc);
    }

    switch (zapytanie.operacja) {
    case (OP_NOWE_POL):
      if (k = dodaj_klienta(argument_const)) {
	write(k, &k, sizeof(int)); /* odpowiedzia jest numer przyznanego lacza, 
					        sklep nie zapamietuje listy otwartych lacz. */
	logowanie("Nowe poczenie: Klient %d\n", k);
      } else {
	logowanie("Bd nowego poczenia\n");
      }
      break;
    case (OP_KONIEC_POL):
      if (close(zapytanie.nadawca)==0)
	logowanie("Klient %d koniec poczenia: Sukces\n", zapytanie.nadawca);
      else
	logowanie ("Klient %d koniec poczenia: Bd\n", zapytanie.nadawca);
      break;
    case (OP_OFERTA):
      logowanie("Klient %d oferta\n", zapytanie.nadawca);
      wypisz_towary(zapytanie.nadawca, towar, l_towarow);
      break;
    case (OP_KUP):
      if (sprzedaj_towar(argument_const, towar, l_towarow)) {
	write(zapytanie.nadawca, "Towar zakupiony\n", 16);
	logowanie("Klient %d zakup dokonany\n", zapytanie.nadawca);
      } else {
	write(zapytanie.nadawca, "Brak towaru/Towar nieznany\n", 27);
	logowanie("Klient %d zakup nie dokonany\n", zapytanie.nadawca);
      }
      break;
    }
  }
}

/* gdy tryb==zapamietaj ustawienia, nie zamyka sklepu, tylko
   zapamietuje co bedzie trzeba zamknac w razie zamykania z bezpieczne_wyjscie(),
   gdzie nie mozna przekazac parametrow, bo to jest signal handler */
void zamknij_sklep(int lacze, towar_t* towar, int ltow, tryb_t tryb) {
  static int ost_lacze, ost_ltow;
  static towar_t* ost_towar;
  int i;
  if (tryb & (zapamietaj_ustawienia|wykonaj_dla_argumentow)) {
    ost_lacze=lacze;
    ost_towar=towar;
    ost_ltow=ltow;
  } 
  if (tryb & (wykonaj|wykonaj_dla_argumentow)) {
    close(ost_lacze);
    unlink(SCIEZKA_LACZA);
    for (i=0; i<ost_ltow; i++)
      free(ost_towar[i].nazwa);
    free(ost_towar);
  }
}

void bezpieczne_wyjscie(int signum) {
  logowanie("Otrzymano sygna wyjcia\n");
  zamknij_sklep(0, NULL, 0, wykonaj); /* */
  exit(0);
}

int main(int argc, char *argv[]) {
  towar_t* magazyn=NULL;
  int lacze, ltow;
  if (argc<2) {
    fprintf(stderr, "Uywanie: sklep plik\n"
	            "\tplik - plik z zawartoci magazynu\n");
    return 1;
  }
  if ((magazyn = laduj_towar(argv[1], &ltow)) == NULL) {
    fprintf(stderr, "Nie mona zaadowa magazynu\n");
    return 2;
  }
  logowanie("Towar zaadowany\n");
  if (! (lacze = otworz_lacze(SCIEZKA_LACZA, O_RDWR))) {
    fprintf(stderr, "Nie mona utworzy cza. Moe dziaa ju inny sklep, "
                                    "lub ostatnio program zamknito niepoprawnie.\n");
    return 3;
  }
  kasa(0, KASA_STAN);
  logowanie("Poczenie otwarte\n");
  signal(SIGINT, bezpieczne_wyjscie);

  oferuj(lacze, magazyn, ltow);
  zamknij_sklep(lacze, magazyn, ltow, wykonaj_dla_argumentow);
  return 0;
}
