#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>

#define PIPE_FOR_READ 0
#define PIPE_FOR_WRITE 1

#define RD_LEFT 0 // pipe for read from left
#define RD_BACK 1 // pipe for read from back
#define RD_UP   2 // pipe for read from up
#define WR_RIGHT   3 // pipe for write to right
#define WR_FORWARD 4 // pipe for write to forward
#define WR_DOWN    5 // pipe for write to down

#define CLNT_NOOPTS 0
#define CLNT_STDIN  1
#define CLNT_STDOUT 2

#define BUFSIZE 1024

typedef int t_pipe[2];
typedef int forkpipes[6]; 

float deform=0.0;

int subnproc_pipe(int n, int *writefd) {
  t_pipe newp;
  if (n==0) return 0;
  pipe(newp);
  if (fork()==0) {
    close(newp[PIPE_FOR_WRITE]);
    subnproc_pipe(n-1, writefd); // rekurencyjny ciag procesow
    transitclient(newp[PIPE_FOR_READ], *writefd);
    exit(0);
  } else {
    *writefd = newp[PIPE_FOR_WRITE];
    close(newp[PIPE_FOR_READ]);
  }
}

char *randdeform(char *s) {
  int i, k=strlen(s);
  for (i=0; i<k; i++)
    if ((s[i]>63)&&((rand()*1.0)/RAND_MAX<deform))
      s[i]=97+(int)(rand()*25.0/(RAND_MAX));
  return s;
}

int is_open_any_of(FILE *f1, FILE *f2, FILE *f3) {
  return (! ((f1==NULL||feof(f1))&&(f2==NULL||feof(f2))&&(f3==NULL||feof(f3))));
}

char *sumstr(char *buf, char** sumup) {
  int i, j, k;
  memset(buf, 0, BUFSIZE);
  k=strlen(sumup[0]);
  for (i=0; i<k; i++) {
    j=sumup[0][i];
    if ((sumup[1][i]!=j)&&(sumup[1][i]==sumup[2][i])) j=sumup[1][i];
    if ((sumup[1][i]!=j)&&(sumup[2][i]!=j)) j='?';
    buf[i]=j;
  }
  return buf;
}

int client(forkpipes fd, int options) {
  int i, j, k=0, bufempty=1, exit=0;
  char *buf, *tmp, **sumup;
  FILE *f[6];
  buf = (char*) malloc(BUFSIZE*sizeof(char));
  if (options==CLNT_STDOUT) {
    sumup = (char**) malloc(3*sizeof(char*));
    for (i=0; i<3; i++)
      sumup[i] = malloc(BUFSIZE*sizeof(char));
  } else
    tmp = (char*) malloc(BUFSIZE*sizeof(char));
    
  for (i=0; i<3; i++)
    f[i] = fd[i] ? fdopen(fd[i], "r") : NULL;
  if (options==CLNT_STDIN)
    f[RD_LEFT]= stdin;

  if (options!=CLNT_STDOUT)
    for (i=3; i<6; i++)
      f[i]= fd[i] ? fdopen(fd[i], "w") : NULL;

  srand((int) getpid()*time(NULL));

  while (is_open_any_of(f[RD_LEFT], f[RD_BACK], f[RD_UP])&&(! exit)) {
    memset(buf, 0, BUFSIZE);
    bufempty=1;
    for (i=0; i<3; i++)
      if (!(f[i]==NULL||feof(f[i]))) {
	if (options!=CLNT_STDOUT) {
	  fgets(tmp, BUFSIZE, f[i]);
	  k=strlen(tmp);
	  if (bufempty) {
	    memcpy(buf, tmp, BUFSIZE); bufempty=0; }
	  else
	    for (j=0; j<k; j++)
	      if ((rand()*1.0)/RAND_MAX>0.3)
		buf[i]=tmp[i];
	}
	else
	  fgets(sumup[i], BUFSIZE, f[i]);
      }

    if (buf[0]=='?') exit=1;

    if (options==CLNT_NOOPTS)
      randdeform(buf);

    if (options==CLNT_STDOUT)
      fputs(sumstr(buf, sumup), stdout);
    else   
      for (i=3; i<6; i++)
	if (fd[i]!=0) { 
	  fputs(buf, f[i]);
	  fflush(f[i]);
	}
  }
  for (i=0; i<6; i++) {
    if (f[i]!=NULL)  fclose(f[i]);
    close(fd[i]);
  }
  free(buf);
  if (options==CLNT_STDOUT) {
    for (i=0; i<3; i++) free(sumup[i]);
    free(sumup);
  }
  else free(tmp);
  return 0;
}

int transitclient(int fromfd, int tofd) {
  forkpipes fd={fromfd, 0, 0, tofd, 0, 0};
  return client(fd, CLNT_NOOPTS);
}

int main(int argc, char *argv[]) {
  int x, y, z, i, subn;
  t_pipe tmppipe;
  forkpipes tmppipes;
  int w=2, h=2, d=2;
  forkpipes ***cube;
  if (argc<3) {printf("forkcube n f\n"
		      "\tn - ilosc procesow na krawedzi\n"
		      "\tf - wspolczynnik przeklaman\n"); exit(1); }
  subn=atoi(argv[1]);
  deform=atof(argv[2]);

  srand((int) time(NULL));
  cube = (forkpipes***) malloc(sizeof(forkpipes**)*d);
  for (i=0; i<h; i++) {
    cube[i] = (forkpipes**) malloc(sizeof(forkpipes*)*h);
    for (x=0; x<w; x++)
      cube[i][x] = (forkpipes*) malloc(sizeof(forkpipes)*w);
  }

  for (z=0; z<d; z++)
    for (y=0; y<h; y++)
      for (x=0; x<w; x++) { // create 3 pipes for each process
	i=RD_LEFT;
	if (x<w-1) {
	  if (pipe(tmppipe)) {fprintf(stderr, "not enough pipes\n"); exit(1);}
	  cube[z][y][x+1][RD_LEFT] = tmppipe[PIPE_FOR_READ];
	  cube[z][y][x][WR_RIGHT] = tmppipe[PIPE_FOR_WRITE]; }
	i=RD_BACK;
	if (z<d-1) {
	  if (pipe(tmppipe)) {fprintf(stderr, "not enough pipes\n"); exit(1);}
	  cube[z+1][y][x][RD_BACK] = tmppipe[PIPE_FOR_READ];
	  cube[z][y][x][WR_FORWARD] = tmppipe[PIPE_FOR_WRITE]; }
	i=RD_UP;
	if (y<h-1) {
	  if (pipe(tmppipe)) {fprintf(stderr, "not enough pipes\n"); exit(1);}
	  cube[z][y+1][x][RD_UP] = tmppipe[PIPE_FOR_READ];
	  cube[z][y][x][WR_DOWN] = tmppipe[PIPE_FOR_WRITE]; }
      }

  for (z=d-1; z>=0; z--)
    for (y=h-1; y>=0; y--)
      for (x=w-1; x>=0; x--) {
	if ((x==0&&y==0&&z==0)||(fork()==0)) {
	  if (x>0) { close(cube[z][y][x-1][WR_RIGHT]); }
	  if (y>0) { close(cube[z][y-1][x][WR_DOWN]); }
	  if (z>0) { close(cube[z-1][y][x][WR_FORWARD]); }
	  if (x<w-1) { int q=cube[z][y][x][WR_RIGHT];
	    close(cube[z][y][x+1][RD_LEFT]); 
	    subnproc_pipe(subn, &cube[z][y][x][WR_RIGHT]); }
	  if (y<h-1) { 
	    close(cube[z][y+1][x][RD_UP]); 
	    subnproc_pipe(subn, &cube[z][y][x][WR_DOWN]); }
	  if (z<d-1) { 
	    close(cube[z+1][y][x][RD_BACK]); 
	    subnproc_pipe(subn, &cube[z][y][x][WR_FORWARD]); }
	  memcpy(&tmppipes, cube[z][y][x], sizeof(forkpipes)); 
	  free(cube);
	  if (z==0&&y==0&&x==0) { // it's a server & reads from stdin
	    client(tmppipes, CLNT_STDIN);
	  } else if (z==d-1&&y==h-1&&x==w-1) { // it's a last client & writes to stdout
	    client(tmppipes, CLNT_STDOUT);
	  } else {                           // do client jobs
	    client(tmppipes, CLNT_NOOPTS);
	  }
	  exit(0);
	}
      }
  return 0;
}
