/*******************************

 slownik T9 program 
 zaliczeniowy z Lab.PN
 
 Jacek Pospychaa inf66301 grI4

 13.06.2004

 kompilacja:
 linux: 
  gcc -lncurses -o t9 t9.c

 windows: 
  tcc t9.c

 have fun!

 *******************************/

#include <stdio.h>		/* printf, fopen */
#include <string.h>		/* memset, memcpy */
#include <time.h>               /* time */
#ifdef __linux__
#include <ncurses.h>		/* ncurses stands for 'new curses' */
#else
#include <conio.h>
#endif

/* domyslny slownik */
#define DEFAULT_DICT "slownik.txt"

/* 
   kody klawiatury 
   KEY_ABC   - kod klawisza z pierwszymi literami (ABC), 
               kolejne klawisze to te o kodach bezposrednio nastepnych
   KEY_QUIT  - kod klawisza wyjscia z programu
   KEY_BCKSP - kod klawisza backspace
   KEY_ADD   - kod klawisza dodania slowa
   KOD_DEL   - kod klawisza usuniecia slowa
*/
#define KEY_ABC   '2'
#define KEY_QUIT  '0'
#ifdef __NCURSES_H
#define KEY_BCKSP 127
#else
#define KEY_BCKSP 8
#endif
#define KEY_ADD   'a'
#define KEY_DEL   'd'

/* czas w sekundach, w ktorym mozna zmieniac literke wciskajac ciagle ten sam klawisz */
#define LETTER_TIMEOUT 2

#define PREFIX_SET   1
#define PREFIX_GET   2
#define PREFIX_CLEAR 4

int left = 0;			/* ile jeszcze slow wyswietlic na ekranie */
int save = 0;                   /* czy zapisywac do pliku (save=1), czy wyswietlac na ekran (save=0) */
FILE* sfile = NULL;             /* plik do ktorego zapisywane sa slowa podczas zapisu */

/*
  organizacja slownika
  drzewo, ktore w kazdym wierzcholku ma liste liter (wskaznik wd_next), od ktorych
  moze sie zaczynac slowo. lista jest upozadkowana alfabetycznie.
   kazda taka litera ma wskaznik do nastepnego wierzcholka
  (wskaznik wd_after), gdzie jest lista z literami nastepujacymi po niej.

  przykladowo: 
  slownik zawierajacy slowa: fee foo fine line lost link li
  ma strukture:

      f-----------------------l-NULL                        -   <=> wd_next
      |                       |                             |   <=> wd_after
      e------i------o-NULL    i------------------o-NULL  [a-z0] <=> wd_char
      |      |      |         |                  |    
      e-NULL n-NULL o-NULL    0----n-NULL        s-NULL    
      |      |      |         |    |             |    
      0-NULL e-NULL 0-NULL    NULL e------k-NULL t-NULL
      |      |      |              |      |      | 
      NULL   0-NULL NULL           0-NULL 0-NULL 0-NULL
             |                     |      |      |
             NULL                  NULL   NULL   NULL
*/

typedef struct word
{
  char wd_char;			/* litera slowa */
  struct word *wd_next;		/* wskaznik do innej litery na tym samym poziomie */
  struct word *wd_after;	/* wskaznik do nastepnych liter w slowie  */
} word_t;



/*
 *  funkcje dict_here_*
 *
 *  manipulacja listami wd_next: 
 *  dodawanie elementu (w porzadku alfabetycznym), usuwanie elementu,
 *  usuwanie wszystkich elementow, wyszukiwanie elementu po znaku (wd_char)
 */

/* jesli c jest na liscie (wd_next), zwraca wskaznik do niego, jesli nie NULL */
word_t *dict_here_char (word_t * list, char c)
{
  while ((list != NULL) && (list->wd_char < c))
    list = list->wd_next;
  
  return ((list != NULL) && (list->wd_char == c)) ? list : NULL;
}

/* dodaje c na liscie (wd_next) o ile go nie ma i zwraca wskaznik do nowo dodanego
   jesli nowy element ma byc na poczatku, wskaznik poczatku sie nie zmienia
   dodaje elementy na liscie w porzadku alfabetycznym
   zwraca wskaznik do nowo wstawionej litery
*/
word_t *dict_here_add (word_t * list, char c)
{
  word_t *tmp = NULL;
  
  if (list == NULL) {	/* jesli lista pusta */
    list = (word_t *) calloc (1, sizeof (word_t));
    list->wd_char = c;
    
    return list;  
  } else {
    while ((list != NULL) && (list->wd_char < c)) { /* znajdz poprzednia i nastepna litere */
      tmp = list;
      list = list->wd_next;
    }
    
    if (list == NULL)		/* jesli na koncu listy */
      list = tmp;
    else if (list->wd_char == c)
      return list;		/* ten element juz istnieje */
    
    tmp = (word_t *) calloc (1, sizeof (word_t));
    tmp->wd_next = list->wd_next;	/* polacz wskazniki */
    list->wd_next = tmp;
    
    if (list->wd_char > c) {  /* zrob tak, zeby najpierw byla mniejsza litera, potem wieksza */
      tmp->wd_char = list->wd_char;
      list->wd_char = c;
      
      tmp->wd_after = list->wd_after;
      list->wd_after = NULL;
      
      return list;
    }
    else {
      tmp->wd_char = c;
      
      return tmp;
    }
  }
}

/* usuwa c z listy (wd_next) i zwraca wskaznik do poczatku, jesli nie ma c, zwraca NULL */
word_t *dict_here_del (word_t * list, char c)
{
  word_t *prev = NULL;
  
  while ((list != NULL) && (list->wd_char != c)) {
    prev = list;
    list = list->wd_next;
  }
  
  if (list == NULL) /* nie znaleziono elementu, lub lista jest pusta */
    return NULL;
  
  else { /* znaleziono element */
    if (prev != NULL) { /* znaleziony element jest w srodku */
      prev->wd_next = list->wd_next;
      free (list);
      
      return list->wd_next;
    } else { /* znaleziony element jest na poczatku */
      if (list->wd_next == NULL) { /* jest on jedyny */
	free (list);
	return NULL;

      } else { /* jest tylko pierwszym */
	prev = list;
	list = list->wd_next;
	prev->wd_char = list->wd_char;
	prev->wd_next = list->wd_next;
	free (list);
	return prev;

      }
    }
  }
}

/* usuwa cala liste wd_next */
void dict_here_delall (word_t * list)
{
  word_t *prev;
  
  while (list != NULL) {
    prev = list->wd_next;
    free (list);
    list = prev;
  }

  return;
}

/*
 * funkcje dict_word*
 *
 * manipulacja w obu wymiarach: wd_next i wd_after
 *
 */

/* nowe slowo do slownika */
word_t *dict_word_add (word_t * dict, char *word)
{
  word_t *tmp;

  if (word == NULL)
    return dict;
  
  if (dict == NULL)
    dict = dict_here_add (dict, word[0]);
  else
    dict_here_add (dict, word[0]);
  
  tmp = dict_here_char (dict, word[0]);
  
  if (word[0] != 0) {
    word++;
    tmp->wd_after = dict_word_add (tmp->wd_after, word);
  }
  
  return dict;
}

word_t *dict_words_last (word_t * dict, char *buf)
{
  word_t *tmp;

  if ((tmp = dict_here_char (dict, buf[0])) != NULL) {
      if (buf[1] == 0)
	return tmp;
      else
	return dict_words_last (tmp->wd_after, ++buf);
  } else
    return NULL;
}

/* usuniecie wyrazu */
word_t *dict_word_del (word_t * dict, char *word)
{
 word_t* node;
 int c;

 if (strlen(word)==0)
   return dict;

 node = dict_words_last(dict, word);
 dict_here_delall(node->wd_after);
 node->wd_after = NULL;

 return dict;
}

/* zwraca wskaznik do miejsca w slowniku, gdzie jest ostatnia litera z ciagu w buforze */
word_t *dict_words_locate (word_t * dict, char *buf)
{
  word_t *tmp;

  if ((tmp = dict_here_char (dict, buf[0])) != NULL) {
    if (buf[1] == 0)
      return tmp->wd_after;
    else
      return dict_words_locate (tmp->wd_after, ++buf);
  } else
    return NULL;
}

/* zwolnienie pamieci slownika */
void dict_destroy (word_t * dict)
{
  word_t *tmp;

  if (dict != NULL) {
    tmp = dict;
    
    while (tmp != NULL) {
      dict_destroy (tmp->wd_after);
      tmp = tmp->wd_next;
    }

    dict_here_delall (dict->wd_next);
    
    free (dict);
  }
}

/* manipulacja prefiksem, zwraca biezacy ciag */
static char *dict_prefixctl (int cmd, char *prefix)
{
  static char *prfx = NULL;

  switch (cmd) {
  case (PREFIX_SET):		/* ustawienie nowego */
    if (prefix == NULL)
      return NULL;
    
    if (prfx == NULL)
      prfx = (char *) malloc (strlen (prefix) + 1);
    else
      prfx = (char *) realloc (prfx, strlen (prefix) + 1);
    
    memcpy (prfx, prefix, strlen (prefix) + 1);	/* +1 = zero konczace stringa */
    break;
    
  case (PREFIX_CLEAR):	/* skasowanie */
    free (prfx);
    prfx = NULL;
    break;
    
  }
  return prfx;
}

/* funkcja realizujaca wyswietlanie slowa ze slownika */
void dict_printw (char *buf)
{
if (save) {
  fprintf(sfile, "%s%s\n", dict_prefixctl (PREFIX_GET, NULL), buf);
  left++;
} else {
#ifdef __NCURSES_H
  printw ("%s%s\n", dict_prefixctl (PREFIX_GET, NULL), buf);
#endif
#ifdef __CONIO_H
  printf ("%s%s\n", dict_prefixctl (PREFIX_GET, NULL), buf);
#endif
}
  return;
}

/* wypisz pierwsze left ciagow ze slownika */
void dict_words_print (word_t * dict, char *buf, int bufsize, int n)
{
  if ((n >= bufsize) || (left <= 0))
    return;
  
  while (dict != NULL) {
    buf[n] = dict->wd_char;
    
    if ((buf[n] == 0) && (left-- > 0))
      dict_printw (buf);
    else
      dict_words_print (dict->wd_after, buf, bufsize, n + 1);
    
    dict = dict->wd_next;
  }
  buf[n] = 0;
}

/* wypisuje wszystkie slowa ze slownika dict zaczynajace sie od ciagu prefix */
void dict_words_print_after (word_t * dict, char *prefix, int max)
{
  char tmp[64] = "";

  dict_prefixctl (PREFIX_SET, prefix);
  dict = dict_words_locate (dict, prefix);
  left = max;
  dict_words_print (dict, tmp, 64, 0);

}

/* w miejscu pierwszego lepszego konca \n linii wstawia koniec ciagu 0 */
void chop (char *buf)
{
  int i = 0;

  while ((buf[i] != 0) && (buf[i] != '\n'))
    i++;
  
  buf[i] = 0;
}

/* laduje slownik z pliku filename */
word_t *load_dict (char *filename)
{
  word_t *dict = NULL;
  char buf[64];
  FILE *f;

  if ((f = fopen (filename, "r")) == NULL) {
    printf ("Nie mozna otworzyc pliku slownika\n");
    return NULL;
  } else {
    while (fgets (buf, 64, f) != NULL) {
      chop (buf);
      dict = dict_word_add (dict, buf);
    }
    return dict;
  }
}

/* zapisuje slownik dict do pliku filename */
void save_dict (word_t* dict, char *filename)
{
  char l[2]={0};
  int i;
  if ((sfile = fopen(filename, "w")) == NULL)
    return;
  
  save=1;
  for (i='a'; i<='z'; i++) {
    l[0]=i;
    dict_words_print_after(dict, l, 10);
  }
  save=0;
  
  fclose(sfile);
}

/* czytanie wejscia, przetwarzanie polecen i wyswietlanie wynikow */
void go_interactive (word_t * dict, int max)
{
  char*   letters[] = {        "abc", "def",
		       "ghi",  "jkl", "mno",
	  	       "pqrs", "tuv", "wxyz"};
  char   buffer[64];
  int    p = 0;
  int    key=-1, lkey=-1, timeout=0;
  time_t reflex=0;

#ifdef __NCURSES_H
  initscr ();
  cbreak ();
  noecho ();
  nonl ();
  intrflush (stdscr, 0);
  printw ("T9+: Wyjscie: %c, nowe slowo: %c, usun slowo: %c, slowa - klawisze 0-9\n ", KEY_QUIT, KEY_ADD, KEY_DEL);
#endif
#ifdef __CONIO_H
  clrscr();
  printf ("T9+: Wyjscie: %c, nowe slowo: %c, usun slowo: %c, slowo - klawisze 0-9\n ", KEY_QUIT, KEY_ADD, KEY_DEL);
#endif

  memset (buffer, 0, 64);

  while (key != KEY_QUIT) {
    key = getch();
    
    if (key == KEY_BCKSP) {
      if (p>0)
	buffer[--p] = 0;
    } else if (key == KEY_ADD) {
      if (p>0)
	dict_word_add(dict, buffer);
    } else if (key == KEY_DEL) {
      if (p>0)
	dict_word_del(dict, buffer);
    } else
      if (((unsigned int)(key - KEY_ABC)<8) && (p<sizeof(buffer)-1)) {
	if ((key == lkey) && ((time(NULL) - reflex) < LETTER_TIMEOUT)) {
	  timeout = (timeout+1) % strlen(letters[(unsigned int)key-KEY_ABC]);
	  p--;
	} else timeout=0;
	buffer[p++] = letters[(unsigned int)key-KEY_ABC][timeout];
	lkey = key;
	reflex = time(NULL);
      }
    
#ifdef __NCURSES_H
      clear ();
      mvprintw (0, 1, buffer);
      move (2, 0);
#endif
#ifdef __CONIO_H
      clrscr ();
      gotoxy(0, 1);
      printf("%s\n", buffer);
      gotoxy(2, 0);
#endif

      dict_words_print_after (dict, buffer, max);
  }
  
#ifdef __NCURSES_H
  endwin ();
#endif
#ifdef __CONIO_H
  clrscr ();
#endif
  return;
}


int main (int argc, char *argv[])
{
  char *dictionary_file = NULL;
  word_t *dict = NULL;
  int i;
  int chbufsize = 20;

  /* inicjalizacja domyslnych ustawien */
  dictionary_file = (char *) malloc (strlen (DEFAULT_DICT) + 1);
  memcpy (dictionary_file, DEFAULT_DICT, strlen (DEFAULT_DICT) + 1);
  
  /* przetwarzanie argumentow z linii polecen */
  for (i=0; i<argc; i++)
    {
      if (strlen(argv[i])==2) {
	switch (argv[i][1])
	  {
	  case 'f':
	    if ((i==argc-1) || (argv[i+1][0]=='0')) {
	      printf("Opcja f wymagag parametru\n");
	      exit(1);
	    }
	    realloc (dictionary_file, strlen (argv[i+1]) + 1);
	    memcpy (dictionary_file, argv[i+1], strlen (argv[i+1]) + 1);
	    break;
	    
	  case 'l':
	    if ((i==argc-1) || (argv[i+1][0]=='0')) {
	      printf("Opcja l wymagag parametru\n");
	      exit(1);
	    }
	    if ((chbufsize = atoi (argv[i+1])) == 0)
	      {
		printf ("Liczba propozycji musi byc liczba >0\n");
		return 1;
	      }
	    break;
	    
	  case 'h':
	    printf ("Skladnia: t9 [-f plik] [-l n] [-h] [-v]\n"
		    "-f plik Nazwa pliku slownika, ktory ma byc uzyty\n"
		    "-l n    Liczba linii podpowiedzi (Domyslnie 20)\n"
		    "-h      Pomoc, ktora widzisz\n"
		    "-v      Informacja o wersji programu\n");
	    return 0;
	    
	  case 'v':
	    printf ("T9+ program zaliczeniowy z PN\n"
		    "Jacek Pospychala (inf66301)\n");
	    return 0;
	    
	  }
      }
    }
  
  /* zaladowanie slownika */
  if ((dict = load_dict (dictionary_file)) == NULL)
    return 1;
  
  /* interfejs */
  go_interactive (dict, chbufsize);
  
  /* zapisanie slownika ze zmianami ktore wprowadzil user */
  save_dict(dict, dictionary_file);
  
  /* posprzatanie po sobie */
  dict_destroy (dict);
  
  return 0;
}
